package lbi.relational;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Predicate;

/** A relational table as a multiset of tuples. */
public class Table {

	/** Table name. */
    private String name;
    
    /** Table tuples. Storage is a list, which allows 
     *  for sets, multisets and ordered list of tuples. 
     *  Checks that tuple have all the same type is omitted. */
	private ArrayList<Tuple> rows;
   
	/** Constructor of empty table. */
	public Table(String name) {
		this.name = name;
		rows = new ArrayList<Tuple>();
	}
	
	/** Constructor with initial list of tuples. */
	public Table(String name, ArrayList<Tuple> rows) {
		this.name = name;
		this.rows = rows;
	}

	/** Return table name. */
	public String getName() {
		return name;
	}
	
	/** Table contents as a string. */
	public String toString() {
		StringBuilder result = new StringBuilder("Table: " + name);
		for(Tuple t : rows)
			result.append("\n" + t);
		return result.toString();		
	}
	
	/** Insert a tuple. */
	public void insert(Tuple t) {
		rows.add(t);
	}
   
	/** Project (multiset) operator. */
	public Table project(String []attributes) {
		ArrayList<Tuple> result = new ArrayList<Tuple>();
		for(Tuple t : rows)
			result.add( t.project(attributes) );
		return new Table(name+".project", result);		
	}
	
	/** Rename operator. */
	public Table rename(Map<String, String> atob) {
		ArrayList<Tuple> result = new ArrayList<Tuple>();
		for(Tuple t : rows)
			result.add( t.rename(atob) );
		return new Table(name+".rename", result);		
	}
	
	/** Distinct tuples operator. */
	public Table distinct() {
		HashSet<Tuple> set = new HashSet<Tuple>(rows);
		ArrayList<Tuple> result = new ArrayList<Tuple>(set);
		return new Table(name+".distinct", result);		
	}

	/** Selection operator. */
	public Table select(Predicate<Tuple> c) {
		ArrayList<Tuple> result = new ArrayList<Tuple>();
		for(Tuple t : rows)
			if(c.test(t))
				result.add(t);
		return new Table(name+".select", result);		
	}
	
	/** Sort operator. */
	public Table sort(String []attributes) {
		ArrayList<Tuple> result = new ArrayList<Tuple>(rows);
		Collections.sort(result, (t1, t2) -> t1.compare(t2,  attributes)); // sort by comparing on attributes
		return new Table(name+".sort", result);		
	}
	
	/** Group by operator. */
	public Table groupby(String A, String B) {
		// hashmap to hold for every value of A the number of tuples with that value
		HashMap<Object, MInteger> hashGroupBy = new HashMap<Object, MInteger>();
		for(Tuple t : rows) { // for every row
			Object value = t.getObject(A); // find value of A 
			MInteger counter = hashGroupBy.get(value); // lookup 
			if(counter == null) { // if not found
				counter = new MInteger(); // counter initial value is 0
				hashGroupBy.put(value, counter); // add to map
			}
			counter.intValue++; // increment counter
		}
		ArrayList<Tuple> result = new ArrayList<Tuple>();
		for( Entry<Object, MInteger> p : hashGroupBy.entrySet()) { // for every value of A
			Tuple t = new Tuple(); // new tuple
			t.setObject(A, p.getKey()); // value of A
			t.setInt(B, p.getValue().intValue); // value of Count(*) AS B
			result.add(t);
		}
		return new Table(name+".groupby", result);		
	}
	
	
	/** Equi-join operator. */
	public Table join(Table other, String A, String B) {
		// hashGroupBy holds for every value of A the list of tuples having such a value
		HashMap<Object, List<Tuple>> hashGroupBy = new HashMap<Object, List<Tuple>>();
		for(Tuple t : rows) { // for every tuple
			Object value = t.getObject(A); // find value of A  
			List<Tuple> list = hashGroupBy.get(value); // find value of A 
			if(list == null) { // if not found
				list = new ArrayList<Tuple>(); // empty list
				hashGroupBy.put(value, list); // add to map
			}
			list.add(t); // add t to the list
		}
		ArrayList<Tuple> result = new ArrayList<Tuple>();
		for(Tuple t : other.rows) { // for every tuple in the right join table
			Object value = t.getObject(B); // find value of B
			List<Tuple> list = hashGroupBy.get(value); // find list of matching tuples in A 
			if(list == null) // if no matching 
				continue; // do noting
			for(Tuple t2 : list) // for every matching tuple
				result.add(t.join(t2)); // join tuples
		}
		return new Table(name+".join", result);		
	}
	
}
