package lbi.relational;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

/** A tuple. */
public class Tuple {
	/** Tuple storage. */
	private HashMap<String, Object> values;
    
	/** Empty constructor. */
 	public Tuple() {
    	values = new HashMap<String, Object>();
    }
 	
    /** Get value of an integer attribute. */
	public int getInt(String attribute) {
		return (int)values.get(attribute);
	}
	
    /** Get value of a string attribute. */
	public String getString(String attribute) {
		return (String)values.get(attribute);
	}
	
    /** Get value of an attribute. */
	public Object getObject(String attribute) {
		return values.get(attribute);
	}

    /** Set value of an integer attribute. */
	public void setInt(String attribute, int value) {
		values.put(attribute, value);
	}
		
    /** Set value of a string attribute. */
	public void setString(String attribute, String value) {
		values.put(attribute, value);
	}

    /** Set value of an attribute. */
	public void setObject(String attribute, Object value) {
		values.put(attribute, value);
	}

	/** Table contents as a string. */
	public String toString() {
		StringBuilder res = new StringBuilder("{ "); // StringBuilder allows append operations
		for(Entry<String, Object> p : values.entrySet()) // for every pair
			res.append("("+p.getKey()+", "+p.getValue()+") ");
		res.append("}");
		return res.toString();
	}

	/** Equality of two tuples. Needed when putting tuples in a set. */
	public boolean equals(Object ob) {
		Tuple o = (Tuple)ob;
		return values.equals(o.values);
	}

	/** Hash-code of a tuples. Needed when putting tuples in a hash set.
	 *  HashSet first locate the bucket using the hash-code, then check
	 *  equality using equals(). */
	public int hashCode() {
		return values.hashCode();
	}
	
	/** Compare to another tuple on an array of attributes. 
	 *  This method is more general than compareTo() of interface Comparable. */
	public int compare(Tuple t1, String []atts) {
		for(String att : atts) {
			@SuppressWarnings({"unchecked" }) // assumes that types of attributes are Comparable (Integer and String are!)
			int c = ((Comparable<Object>)values.get(att)).compareTo(t1.values.get(att)); 
			if(c != 0) // compare values of attributes in atts
				return c;
		}
		return 0;		
	}
	
	/** Project a single tuple. */
	public Tuple project(String[] attributes) {
		Tuple result = new Tuple();
		for(String a : attributes)
			result.values.put(a, values.get(a));
		return result;
	}
	
	/** Rename attributes a single tuple. */
	public Tuple rename(Map<String, String> atob) {
		Tuple result = new Tuple();
		for(Entry<String, Object> p : values.entrySet()) {
			String att = p.getKey();
			String rename = atob.get(att);
			if(rename!=null)
				att = rename;
			result.values.put(att, p.getValue());
		}
		return result;
	}
	
	/** Join with another tuple. */
	public Tuple join(Tuple t2) {
		Tuple result = new Tuple();
		for(Entry<String, Object> p : values.entrySet()) 
			result.values.put(p.getKey(), p.getValue());
		for(Entry<String, Object> p : t2.values.entrySet()) 
			result.values.put(p.getKey(), p.getValue());
		return result;
	}
}
